//---------------------------------------------------------------------------
//
// $Workfile:   ESRemap.cpp  $
// $Version$
// $Version$   1.0
// Description:  A utility for mapping 24-bit bmp's to 8-bit bmps using a phoenix palette
//
// Initial revision. Louie McCrady
//
//---------------------------------------------------------------------------

#include <sim.h>
#include <dir.h>
#include <g_bitmap.h>
#include <stringtable.h>
#include <commctrl.h>

StringTable stringTable;

#define ESTIMATE    "%.1f seconds"
#define CURRENT     "%s"
#define CURRENT_IPL "%s"

#define STATIC_ESTIMATE    "Time Left:"
#define STATIC_CURRENT     "Current BMP:"
#define STATIC_CURRENT_IPL "Current IPL:"

static const char *szAppName   = "Starsiege Artwork Remapper";
static const char *szClassName = "SSRemapper";
static const char paletteName[MAX_PATH]  = {0,};
static VectorPtr<char*> palettes;
static VectorPtr<char*> bitmaps;
static HINSTANCE hInstance = NULL;
static HWND hWnd = NULL;
static HWND hProgress = NULL;
static HWND hEstimate = NULL;
static HWND hCurrent = NULL;
static HWND hButton = NULL;
static HWND hPalette = NULL;
static HWND hStatic1 = NULL;
static HWND hStatic2 = NULL;
static HWND hStatic3 = NULL;
static HFONT hFont1 = NULL;
static HFONT hFont2 = NULL;
static float startTime = 0;
static float totalBitmaps;
static float currentBitmap;
static float progress;
static bool done = false;
static GFXPalette *palette = NULL;

//------------------------------------------------------------------------------
bool yield()
{
   MSG msg;

   if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) 
   {  
      // Check for exit. (ignore, let main loop handle it)
      if (msg.message == WM_QUIT) 
      {
         return false;
      }

      if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) 
      {
         // Normal translate/dispatch
         TranslateMessage(&msg);
         DispatchMessage (&msg);
      }
   }

   if (totalBitmaps)
   {
      float elapse = GetTickCount()-startTime;
      float estimate = totalBitmaps*elapse/(currentBitmap+progress);
      estimate -= elapse;
      estimate /= 1000.0f;    // ms to seconds
      SendMessage(hEstimate, WM_SETTEXT, 0, (LPARAM)avar(ESTIMATE, estimate));
   }

   return (true);
}   

void cleanup()
{
   if (hFont1)
   {
      DeleteObject(hFont1);
   }
   if (hFont2)
   {
      DeleteObject(hFont2);
   }

   while (palettes.size())
   {
      delete [] palettes[0];
      palettes.pop_front();
   }

   while (bitmaps.size())
   {
      delete [] bitmaps[0];
      bitmaps.pop_front();
   }
}

//------------------------------------------------------------------------------
bool findBitmapsForPalette(const char *lpszPalette)
{
   WIN32_FIND_DATA  data;
   HANDLE           hFind;
   char            *lpszTok;
   char             szDir[256];
   char             szBuf[512];

   if (lpszPalette)
   {
      strcpy(szBuf, lpszPalette);

      if ((lpszTok = strrchr(szBuf, '.')) != NULL)
      {
         *lpszTok = '\0';

         strcpy (szDir, szBuf);
         sprintf(szBuf, "%s\\*.bmp", szDir);

         hFind = FindFirstFile(szBuf, &data);

         if (hFind != INVALID_HANDLE_VALUE)
         {
            do
            {
               sprintf(szBuf, "%s\\%s", szDir, data.cFileName);
               bitmaps.push_back(strnew(szBuf));
            } while (FindNextFile(hFind, &data));

            FindClose(hFind);
         }

         SendMessage(hProgress, 
            PBM_SETRANGE, 0, MAKELPARAM(0, bitmaps.size())); 
         SendMessage(hProgress, PBM_SETSTEP, (WPARAM)1, 0); 
         SendMessage(hProgress, PBM_SETPOS, (WPARAM)0, 0);

         UpdateWindow(hWnd);
         totalBitmaps = bitmaps.size();
      }
   }

   return (bitmaps.size() != 0);
}   


//------------------------------------------------------------------------------
bool findPalettes()
{
   WIN32_FIND_DATA data;
   HANDLE          hFind;
   
   hFind = FindFirstFile("*.ipl", &data);

   if (hFind != INVALID_HANDLE_VALUE)
   {
      do 
      {
         palettes.push_back(strnew(data.cFileName));
      } while (FindNextFile(hFind, &data));

      FindClose(hFind);
   }

   return (palettes.size() != 0);
}   


//------------------------------------------------------------------------------
bool processBitmaps()
{
   GFXBitmap  *bitmap, *eightBit;

   static bool message = false;

   // If we have finished all of this palette's bitmaps, 
   // go to the next one
   if (bitmaps.size() == 0)
   {
      // If we're out of palettes, return false
      if (palettes.size() == 0)
      {
         return (false);
      }

      // Otherwise, find the bitmaps for this palette
      findBitmapsForPalette(palettes[0]);

      // Update the window
      SendMessage(hPalette,
         WM_SETTEXT, 0, (LPARAM)avar(CURRENT_IPL, palettes[0]));
      UpdateWindow(hWnd);

      if (palette)
         delete palette;

      palette = GFXPalette::load(palettes[0]);
      
      delete [] palettes[0];
      palettes.pop_front();
   }

   if (palette && bitmaps.size())
   {
      palette->allowedColorMatches.set(0, 254);

      // Update the window
      SendMessage(hCurrent, 
         WM_SETTEXT, 0, (LPARAM)avar(CURRENT, bitmaps[0]));
      UpdateWindow(hWnd);

      bitmap = GFXBitmap::load(bitmaps[0]);
   
      if (bitmap && bitmap->bitDepth == 24)
      {
         eightBit = GFXBitmap::create(bitmap->width, bitmap->height);

         if (eightBit)
         {
            BYTE *src, *dst, *tmp_src, *tmp_dst, r, g, b;

            src = bitmap->pBits;
            dst = eightBit->pBits;

            eightBit->paletteIndex = palette->palette[0].paletteIndex;
            eightBit->pPalette     = palette;

            for (int i = 0, c = 0; i < bitmap->height; i ++)
            {
               tmp_src = src;
               tmp_dst = dst;

               for (int j = 0; j < bitmap->width; j ++)
               {
                  b = *tmp_src ++;
                  g = *tmp_src ++;
                  r = *tmp_src ++;

                  *tmp_dst ++ = palette->GetNearestColor(r, g, b);

                  if (++c && !(c % 1024))
                  {
                     progress = float(c) / float(bitmap->imageSize / 3);

                     if(!yield())
                     {
                        return (true);
                     }
                  }
               }

               src += bitmap->stride;
               dst += eightBit->stride;
            }

            if (!eightBit->write(bitmaps[0]))
            {
               if(MessageBox(0, 
                  avar("Can't write file: %s (read only?)", bitmaps[0]), 
                  NULL, MB_OKCANCEL | MB_ICONASTERISK) == IDCANCEL)
               {
                  while (bitmaps.size())
                  {
                     SendMessage(hProgress, PBM_STEPIT, 0, 0); 
                     delete [] bitmaps[0];
                     bitmaps.pop_front();
                  }

                  UpdateWindow(hWnd);
                  return (true);   
               }
            }

            currentBitmap += 1.0f;
         }
         else
         {
            totalBitmaps -= 1.0f;
         }

         eightBit->pPalette = NULL;  // We don't want this getting deleted
         delete eightBit;
      }
      delete bitmap;
      delete bitmaps[0];
      bitmaps.pop_front();
   }

   SendMessage(hProgress, PBM_STEPIT, 0, 0); 
   UpdateWindow(hWnd);

   if (message == false && palettes.size() == 0 && bitmaps.size() == 0)
   {
      message = true;

      SendMessage(hProgress, PBM_SETPOS, (WPARAM)0, 0); 
      SendMessage(hEstimate, 
         WM_SETTEXT, 0, (LPARAM)avar(ESTIMATE, 0));
      SendMessage(hCurrent,  
         WM_SETTEXT, 0, (LPARAM)avar(CURRENT,     "Finished"));
      SendMessage(hPalette,  
         WM_SETTEXT, 0, (LPARAM)avar(CURRENT_IPL, "Finished"));

      UpdateWindow(hWnd);
   }

   return (bitmaps.size() == 0);
}   


//------------------------------------------------------------------------------
bool initWindow(HWND hWnd)
{
   SIZE  size;
   HDC   hDC = GetDC(hWnd);
   LONG  lMaxSize = 0L;
   RECT  rcClient;
   int   cyVScroll = GetSystemMetrics(SM_CYVSCROLL); 

   GetClientRect(hWnd, &rcClient);        

   // Create fonts
   hFont1 = CreateFont(12,                       // logical height
                       0,                        // logical avg char width
                       0,                        // angle of escapement
                       0,                        // base-line orientation angle
                       FW_BOLD,                  // font weight
                       0,                        // italic attribute flag
                       0,                        // underline attribute flag
                       0,                        // strikeout attribute flag
                       ANSI_CHARSET,             // character set ID
                       OUT_DEFAULT_PRECIS,       // output precision
                       CLIP_DEFAULT_PRECIS,      // clipping precision
                       DEFAULT_QUALITY,          // output quality
                       DEFAULT_PITCH | FF_SWISS, // pitch and family
                       "MS SANS SERIF");         // typeface name

   hProgress = CreateWindowEx(0, 
                              PROGRESS_CLASS, 
                              (LPSTR)NULL, 
                              WS_CHILD | WS_VISIBLE, 
                              rcClient.left, 
                              rcClient.bottom - cyVScroll,         
                              rcClient.right, cyVScroll, 
                              hWnd, 
                              (HMENU)0, 
                              hInstance, 
                              NULL);

   hStatic1  = CreateWindow("STATIC",
                             STATIC_CURRENT_IPL,
                             WS_CHILD | WS_VISIBLE | SS_LEFTNOWORDWRAP,
                             0, 0, 0, 0,
                             hWnd,
                             0, 
                             hInstance,
                             NULL);

   // If the font was created, set it
   if (hFont1)
   {
      SendMessage(hStatic1, WM_SETFONT, (WPARAM)hFont1, MAKELPARAM(TRUE, 0));
   }

   // See how long this window's text is
   GetTextExtentPoint32(hDC, 
         STATIC_CURRENT_IPL, strlen(STATIC_CURRENT_IPL), &size);

   // If it's longer than the current max, re-set the max, but
   // cap it to the window width
   lMaxSize = min(rcClient.right - 75, max(size.cx, lMaxSize));

   // Resize the window
   MoveWindow(hStatic1, rcClient.left + 10, 5, lMaxSize, 15, TRUE);

   hStatic2  = CreateWindow("STATIC",
                             STATIC_CURRENT,
                             WS_CHILD | WS_VISIBLE | SS_LEFTNOWORDWRAP,
                             0, 0, 0, 0,
                             hWnd,
                             0, 
                             hInstance,
                             NULL);

   // If the font was created, set it
   if (hFont1)
   {
      SendMessage(hStatic2, WM_SETFONT, (WPARAM)hFont1, MAKELPARAM(TRUE, 0));
   }

   // See how long this window's text is
   GetTextExtentPoint32(hDC, 
         STATIC_CURRENT, strlen(STATIC_CURRENT), &size);

   // If it's longer than the current max, re-set the max, but
   // cap it to the window width
   lMaxSize = min(rcClient.right - 75, max(size.cx, lMaxSize));

   // Resize the window
   MoveWindow(hStatic2, rcClient.left + 10, 22, lMaxSize, 15, TRUE);

   hStatic3  = CreateWindow("STATIC",
                             STATIC_ESTIMATE,
                             WS_CHILD | WS_VISIBLE | SS_LEFTNOWORDWRAP,
                             0, 0, 0, 0,
                             hWnd,
                             0, 
                             hInstance,
                             NULL);

   // If the font was created, set it
   if (hFont1)
   {
      SendMessage(hStatic3, WM_SETFONT, (WPARAM)hFont1, MAKELPARAM(TRUE, 0));
   }

   // See how long this window's text is
   GetTextExtentPoint32(hDC, 
         STATIC_ESTIMATE, strlen(STATIC_ESTIMATE), &size);

   // If it's longer than the current max, re-set the max, but
   // cap it to the window width
   lMaxSize = min(rcClient.right - 75, max(size.cx, lMaxSize));

   // Resize the window
   MoveWindow(hStatic3, rcClient.left + 10, 39, lMaxSize, 15, TRUE);

   hButton = CreateWindow("BUTTON", 
                          "Quit",
                          WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON,
                          rcClient.right - 70, 20,
                          60, 30,
                          hWnd, 
                          0, 
                          hInstance, 
                          NULL);

   if (hFont1)
   {
      // Set the font
      SendMessage(hButton,  WM_SETFONT, (WPARAM)hFont1, MAKELPARAM(TRUE, 0));
   }

   // Now the dynamic text windows
   hFont2 = CreateFont(12,                       // logical height
                       0,                        // logical avg char width
                       0,                        // angle of escapement
                       0,                        // base-line orientation angle
                       FW_THIN,                  // font weight
                       0,                        // italic attribute flag
                       TRUE,                     // underline attribute flag
                       0,                        // strikeout attribute flag
                       ANSI_CHARSET,             // character set ID
                       OUT_DEFAULT_PRECIS,       // output precision
                       CLIP_DEFAULT_PRECIS,      // clipping precision
                       DEFAULT_QUALITY,          // output quality
                       DEFAULT_PITCH | FF_SWISS, // pitch and family
                       "MS SANS SERIF");         // typeface name

   hPalette  = CreateWindow("STATIC",
                            avar(CURRENT_IPL, "No palette"),
                            WS_CHILD | WS_VISIBLE | SS_LEFTNOWORDWRAP,
                            rcClient.left  + 10 + lMaxSize,       5,
                            rcClient.right - 10 - lMaxSize - 70, 15,
                            hWnd,
                            0, 
                            hInstance,
                            NULL);

   hCurrent = CreateWindow("STATIC", 
                           avar(CURRENT, "No bitmap"), 
                           WS_CHILD | WS_VISIBLE | SS_LEFTNOWORDWRAP, 
                           rcClient.left  + 10 + lMaxSize,      22,
                           rcClient.right - 10 - lMaxSize - 70, 15,
                           hWnd, 
                           0, 
                           hInstance, 
                           NULL);

   hEstimate = CreateWindow("STATIC", 
                            avar(ESTIMATE, 0.0f), 
                            WS_CHILD | WS_VISIBLE | SS_LEFTNOWORDWRAP, 
                            rcClient.left  + 10 + lMaxSize,      39,
                            rcClient.right - 10 - lMaxSize - 70, 15,
                            hWnd, 
                            0, 
                            hInstance, 
                            NULL);

   if (hFont2)
   {
      // Set the font
      SendMessage(hPalette,  WM_SETFONT, (WPARAM)hFont2, MAKELPARAM(TRUE, 0));
      SendMessage(hEstimate, WM_SETFONT, (WPARAM)hFont2, MAKELPARAM(TRUE, 0));
      SendMessage(hCurrent,  WM_SETFONT, (WPARAM)hFont2, MAKELPARAM(TRUE, 0));
   }

   UpdateWindow(hWnd);
   return (true);
}   


//------------------------------------------------------------------------------
long FAR PASCAL WndProc( HWND hWnd, unsigned message, WPARAM wParam, LPARAM lParam)
{
   switch (message)
   {
      case WM_CREATE:
      {
         startTime = GetTickCount();
         
         if (!initWindow(hWnd)) 
         {
            return (-1);
         }
         break;
      }

      case WM_COMMAND:
      {
         if ((HWND)lParam == hButton)   
         {
            done = true;
         }
         break;
      }

      case WM_DESTROY:
      {
         PostQuitMessage(0);
         break;
      }
   
      default:
         return DefWindowProc(hWnd, message, wParam, lParam);
   }
   return (0L);
}   


//------------------------------------------------------------------------------
bool initClass( HINSTANCE hInst )
{
   WNDCLASSEX wc;
   hInstance = hInst;
   wc.cbSize         = sizeof(WNDCLASSEX);
   wc.style          = CS_HREDRAW | CS_VREDRAW;
   wc.lpfnWndProc    = WndProc;
   wc.cbClsExtra     = 0;
   wc.cbWndExtra     = 0;
   wc.hInstance      = hInstance;
   wc.hIcon          = LoadIcon( hInstance, MAKEINTRESOURCE(1));
   wc.hCursor        = LoadCursor( NULL, IDC_ARROW );
   wc.hbrBackground  = (HBRUSH)(COLOR_WINDOW);
   wc.lpszMenuName   = NULL;
   wc.lpszClassName  = szClassName;
   wc.hIconSm        = LoadIcon( hInstance, MAKEINTRESOURCE(2));

   return RegisterClassEx( &wc );
}


//------------------------------------------------------------------------------
int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE, LPSTR lpCmdLine, int nCmdShow)
{
   MSG   msg;

   UNREFERENCED_PARAMETER(lpCmdLine);
   UNREFERENCED_PARAMETER(nCmdShow);

   // We must call this to ensure the common controls are setup for
   // this application
   InitCommonControls();

   if (!initClass(hInstance))
      return (false);

   hWnd = CreateWindow(
               szClassName,
               szAppName,
               WS_OVERLAPPEDWINDOW,
               CW_USEDEFAULT,    
               CW_USEDEFAULT,    
               450, 110,     
               NULL,             // hWnd parent   
               NULL,             // hMenu
               hInstance,        // this instance
               NULL              
               );
   if (!hWnd) 
   {
      return (false);
   }

   ShowWindow(hWnd, SW_SHOW);

   if (!findPalettes())
   {
      MessageBox(hWnd, 
         "Unable to locate required .IPL palettes", NULL, MB_OK | MB_ICONSTOP);
      return (false);
   }

   while (!done)
   {
      if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) 
      {
         // Check for exit.
         if (msg.message == WM_QUIT) 
         {
            // Must call cleanup first to free memory
            cleanup();

            return (false);
         }

         // Normal translate/dispatch
         TranslateMessage(&msg);
         DispatchMessage (&msg);
      }
      processBitmaps();
   }

   return (true);   
}   

